
/**
 * @file   genlp.c
 * @author David Bremner
 * @date   Thu Apr 11 15:37:35 2002
 * 
 * @brief  Generate an (integer) linear program for the feasible 
 * matroid polytopes for given info. There are several caveats
 * 
 */

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include "parser.h"
#include "gparams.h"
#include "combinat.h"
#include "oms_error.h"
#include "facet_list.h"
#include "signed_index.h"
#include "basis_equation.h"
#include "eqout.h"
#include "path.h"
#include "pluecker.h"
#include "boundary.h"
#include "nu_int_to_symbol.h"

/** 
   The idea here is to take the convex hull of the feasible 
   0/1 vectors in in R^6 corresponding to valid points 
   for the pluecker equations, and then lift those onto 
   each (appropriate) 6 set of variables.
   
*/
   
#define XOR_ROWS 16

/* This next is straight from lrs. In particular this means the 
 lhs coefficients are negated.*/
static int coeff[XOR_ROWS][7]={
  { 4,-1,-1,1,-1,-1,-1},
  { 4,-1,-1,-1,1,-1,-1},
  { 3,-1,1,-1,-1,-1,1},
  { 3,1,-1,-1,-1,-1,1},
  { 3,-1,1,-1,-1,1,-1},
  { 3,1,-1,-1,-1,1,-1},
  { 2,-1,-1,1,-1,1,1},
  { 1,1,-1,1,1,1,-1},
  { 1,-1,1,1,1,1,-1},
  { 0,1,1,1,-1,1,1},
  { 2,1,1,1,-1,-1,-1},
  { 0,1,1,-1,1,1,1},
  { 2,1,1,-1,1,-1,-1},
  { 2,-1,-1,-1,1,1,1},
  { 1,1,-1,1,1,-1,1},
  { 1,-1,1,1,1,-1,1}
};

static   be_block_p pluecker,conv_eq;
   
static void usage(char ** argv){
  fprintf(stderr,"usage: %s: [options]\n",argv[0]);
  fprintf(stderr,"\nInput/Output:\n");
  fprintf(stderr,"-i <file>\ttake input from <file>\n");
  fprintf(stderr,"-o <file>\twrite input to <file>\n");
  fprintf(stderr,"-f [ilco]\tOutput format\n");
  fprintf(stderr,"\ti\tine format for lrs/cdd/pd\n");
  fprintf(stderr,"\tl\tlpsolve format\n");
  fprintf(stderr,"\tc\tcplex format\n");
}

static void header(eq_format_t form,int variables,
		   int equations, int inequalities){
  int j;

  switch(form){
  case TEST:
    break;
  case INE:
    printf("* output from genlp\n");
    printf("linearity %d", equations);
    for (j=1; j<=equations; j++){
      printf(" %d",j);
    }
    printf("\nbegin\n");
    printf("%d %d integer\n", equations+inequalities, variables+1);
    break;
  case CPLEX:
  case LPSOLVE:
  case OPBDP:
    eq_zero();
    for (j=0; j<variables; j++){
      eq_coeff(j,1);
    }
    eq_type(OBJ_MAX);
    eq_output();
	    
    if (form==CPLEX)
      printf("subject to\n");

    break;

  default:
    fprintf(stderr,"Output format not implemented");
    exit(1);
  } 

}


void cube_bounds(int outform,int var){

  eq_zero();
  eq_coeff(var,1);
  
  if (outform != LPSOLVE){
    eq_rhs(0);
    eq_type(GEQ);
    eq_output();
  }

  eq_rhs(1);
  eq_type(LEQ);
  eq_output();

}
static void footer(eq_format_t form,int bases){
  int j;
  switch(form){
  case TEST:
    break;
  case CPLEX:
  case LPSOLVE:
    printf("int\n");
    for (j=0; j<bases; j++){
      printf(" x%d\n",j);
    }
    if (form==LPSOLVE)
      printf(";");
    else
      printf("end\n");
    break;
  case INE:
    printf("\nend\n");
    break;
  default:
    fprintf(stderr,"Output format not implemented");
    exit(1);
  }

}

/** 
 * Generate equalities for this facet to be on the boundary.
 * 
 * @param form output format
 * @param conv_eq  coefficients for facet equations
 * @param row  which facet index to set.
 */
static void set_facet(be_block_p conv_eq, int row){
  int i;

  eq_zero();
  
  eq_type(EQ);


  for (i=1; i< be_cols(conv_eq); i++){
    eq_zero();
    if (be_sign(conv_eq,row,i) != be_sign(conv_eq,row,0) ){
      eq_rhs(1);
      eq_coeff(be_bindex(conv_eq,row,0),1);    
    } else {
      eq_rhs(0);
      eq_coeff(be_bindex(conv_eq,row,0),-1);    
    }
    

    eq_coeff(be_bindex(conv_eq,row,i),1);
    eq_output();
  }
}

/** 
 * Generate inequalities forcing this facet off the boundary
 * 
 * @param conv_eq 
 * @param row 
 */
static void forbid_facet( be_block_p conv_eq, int row){
  int i,negs=0;

  eq_zero();



  for (i=0; i< be_cols(conv_eq); i++){
    if (be_sign(conv_eq,row,i)==1){
      eq_coeff(be_bindex(conv_eq,row,i),1);
    } else {
      eq_coeff(be_bindex(conv_eq,row,i),-1);
      negs++;
    }
  }
  eq_rhs(be_cols(conv_eq)-1-negs);
  eq_type(LEQ);
  eq_output();
  
  eq_rhs(1-negs);
  eq_type(GEQ);
  eq_output();

}

/** 
 * Generate inequalities enforcing one row of pluecker table
 * 
 * @param pluecker 
 * @param row 
 */
static void pluecker_ieq(be_block_p pluecker, int row){
  int i,j;


  for (i=0; i<  XOR_ROWS; i++){
    eq_rhs(coeff[i][0]);
    eq_zero();
    eq_type(LEQ);
    for (j=0; j<be_cols(pluecker); j++){    
      eq_coeff(be_bindex(pluecker,row,j),-1*coeff[i][j+1]);
    }
    eq_output();
  }
}

static void path2ieq(int *path, int len){
  int i,j;
  int negs=0;
  int *last_sign;
  int group_sign=1;

  
  /* we are going to ignore the first element in the path on the 
     assumption that it is the source */
  path++;
  len--;

  last_sign=calloc(be_bases(conv_eq)+1,sizeof(int));

  eq_zero();


  for (i=0; i< len; i++){

    /* pass 1, decide how the inequalities for this facet are coupled to 
       the previous one 
    */
    for (j=0; j< be_cols(conv_eq); j++){
      int b=be_bindex(conv_eq,path[i],j);
      int sigma=be_sign(conv_eq,path[i],j);
      if (sigma==-1*last_sign[b]){
	group_sign *=-1;
      }
      last_sign[b]=sigma;
    }

    for (j=0; j< be_cols(conv_eq); j++){
      if (group_sign*be_sign(conv_eq,path[i],j)==1){
	eq_coeff_add(be_bindex(conv_eq,path[i],j),1);
      } else {
	eq_coeff_add(be_bindex(conv_eq,path[i],j),-1);
	negs++;
      }
    }
  }

  eq_rhs(len*(be_cols(conv_eq))-1-negs);

  eq_type(LEQ);
  eq_output();

  eq_rhs(1-negs);
  eq_type(GEQ);
  eq_output();
  free(last_sign);
}


int main(int argc, char** argv){
  int c;
  int n,r;
  int num_bases,num_eqns;


  int row,i;
  eq_format_t outform=CPLEX;
  int *bd_eqn;
  int *path;
  int facets;
  facet_p F;
  be_block_p pivots;
  sarray_p found;

  int eqn_count;
  
  params_p params=malloc(sizeof(params_t));
  
  while((c=getopt(argc,argv,"i:o:f:"))!=-1){
    switch(c){
    case 'i':
      freopen(optarg,"r",stdin);
      assert(stdin);
      break;
    case 'o':
      freopen(optarg,"w",stdout);
      assert(stdout);
      break;
    case 'f':
      outform=eq_arg(optarg);
      break;
    default:
      usage(argv);
      exit(1);

    }
  }

  params->uniform=TRUE;

  /* added as a quick workaround to stop crash. Really the profiling
   * call should be disabled
   */
  prof_init(params);
  parse_input(params);

  n=params->n;
  r=params->r;

  num_eqns=binomial(n,r+2)*binomial(r+2,4); 
  num_bases=binomial(n,r);
  facets=binomial(n,r-1);

  eq_init(outform,num_bases);

  pluecker=be_alloc(num_eqns,6,num_bases);
  pl_init(pluecker,n,r);

  conv_eq=be_alloc(binomial(n,r-1),n-r+1,num_bases); 
  be_boundary_init(conv_eq,n,r,params->source_facet,params->sink_facet,
		    params->diameter);



  bd_eqn=calloc(fl_length(parser.facets),sizeof(int));
  TRAP_NULL(bd_eqn);

  eqn_count=0;
  for (F=fl_first(parser.facets) ; F; F=fl_next(F)) {
    if (fl_sign(F)>0){
      bd_eqn[eqn_count]=fl_bindex(F);
      eqn_count++;
    }
  }

  
  header(outform, num_bases,
	 eqn_count*be_cols(conv_eq)+parser.chirotope_count,
	 fl_length(parser.facets)-eqn_count+ XOR_ROWS*be_rows(pluecker)
		+ 2*num_bases);
	 

  eq_set("fe");

  for (i=0; i<eqn_count; i++)
    set_facet(conv_eq,bd_eqn[i]);

  eq_set("ch");

  for (i=0; i <nss_count(parser.basis_signs); i++){
    int beta;
    nu_sign_t sigma;

    nu_signed_index_t si=nss_next(parser.basis_signs);
    
    beta=nsi_index(si);
    sigma=(1+sign2int(nsi_sign(si)))/2;
    
    eq_zero();
    
    eq_type(EQ);
    eq_rhs(sigma);
    eq_coeff(beta,1);
    eq_output();
  }


  eq_set("fi");

  for (F=fl_first(parser.facets) ; F; F=fl_next(F)){
    if (fl_sign(F) < 0 ) {
      forbid_facet(conv_eq,fl_bindex(F));
    }
  }

  eq_set("c");

  for (row=0; row< num_eqns; row++){
    pluecker_ieq(pluecker,row);
  }


  pivots=be_alloc(facets,(n-r+1)*(r-1),
		  facets);
  be_neighbour_init(pivots,n,r);

  found=sa_new(facets,facets);
  for(i=0; i<sa_size(found); i++)
    sa_init(found,i,0);

  path=calloc(params->diameter+1,sizeof(int));
  for(i=0; i<=params->diameter; i++)
    path[i]=-1;

  path[0]=params->source_facet;
  eq_set("p");
  if (params->diameter != -1)
    path_bt(path,0,params->diameter-1,
	    params->sink_facet,pivots,found,path2ieq);

  eq_set("b");
  if (outform != OPBDP){
    for (i=0; i<num_bases; i++)
      cube_bounds(outform,i);
    footer(outform,num_bases);
  }




  return(0);
}
