
/* $Id$ */
/**
 * @file   nuoms.c
 * @author David Bremner <bremner@unb.ca>
 * $Date$
 * 
 * @brief  The main program for \em oms a recursive backtracking chirotope completion program. nonuniform version
 * 
 * 
 */

#include <unistd.h>
#include <stdio.h>

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif


#include "basis_equation.h"
#include "checkpoint.h"
#include "combinat.h"
#include "combinat.h"
#include "common_startup.h"
#include "debug.h"
#include "globaldef.h"
#include "int_queue.h"
#include "int_stack.h"
#include "matroid.h"
#include "nu_boundary.h"
#include "nu_convexity.h"
#include "nu_matroid.h"
#include "nu_pluecker.h"
#include "nu_sign_set.h"
#include "nu_variable_choice.h"
#include "nuoms.h"
#include "oms_opt.h"
#include "output.h"
#include "profile.h"
#include "ridge_graph.h"
#include "sa16_array.h"
#include "shutdown.h"
#include "signed_set.h"
#include "statearray.h"



#ifdef USE_ZRAM
#include <z_checkpoint.h>
#endif

int oms_procsetnum;
int oms_NodeTag, oms_GlobalTag;


/** 
 * Entry point, either from ZRAM or direct.
 * 
 * @param argc number command line arguments
 * @param argv the arguments
 * @param envp the environment
 * 
 * @return 0 on success
 */
int MAIN(int argc, char *argv[], char **envp){

  na_node_p local,global;
  params_p params;
  
  startopt_t startopt=SO_backtrack|SO_pluecker|SO_nonuniform;

  prof_init();

  global=cs_init_global(argc,argv,startopt);

  params=na_access(global,SLOT_params);

  output_init(params);
  if (params->restart){
#ifdef USE_ZRAM    
    (void)z_bt_RestartDFExp(oms_procsetnum, 
			    (z_bt_GlobalP) global,params->sequential);
#else
      err_fatal(HERE,"Checkpointing only supported with ZRAM");
#endif
  } else {
    do{
      int beta0;
      be_block_p matroid_eq;
      sarray_p matroid_zero_mask,chirotope;

      /* it is not so clear that these should be inside this loop ; 
	 there is probably too much allocation and freeing going on
       */
      local=cs_alloc_local(global,startopt);
      local=nuoms_alloc_local(global,local);
      
      local=cs_init_local(global,local,startopt,nuoms_read_boundary_hook);

      if (!local) 
	shutdown(global,HERE,"inconsistent input");

      params=na_access(global,SLOT_params);
      params->success_count=0;
      matroid_eq=na_access(global,SLOT_matroid_eq);
      matroid_zero_mask=na_access(local,SLOT_matroid_zero_mask);
      chirotope=na_access(local,SLOT_chirotope);

      err_debug(DB_STARTUP,
		(HERE,"initial chirotope: %s",string_pm(chirotope,sign2char)));

      for (beta0=params->first_root_basis; 
	   beta0 <=params->last_root_basis
	     && params->success_count<=0; 
	   beta0++){
	nu_sign_t sigma0=sa_read(chirotope,beta0);
	lparams_p lparams=na_access(local,SLOT_lparams);
	if (sign_intersection(SIGN_NONZERO,sigma0)!=SIGN_NONE){
	  SaveState(local);
	  err_debug(DB_BACKTRACK,("nuoms","trying root basis %d",beta0));
	  if (!matroid_eq->initialized)
	    me_init(matroid_eq,matroid_zero_mask,params->n,params->r,beta0);
	  lparams->root_basis=beta0;
	  if (SetAndConsequences(global,local,beta0,SIGN_NONZERO)){

#ifdef USE_ZRAM
	    if (params->sequential){
	      z_bt_EnumerateDepthFirst(oms_procsetnum, (z_bt_GlobalP) global,
				       (z_NodeP)local);
	    } else {
	      z_bt_EnumerateDFExp(oms_procsetnum, (z_bt_GlobalP) global,
				  (z_NodeP)local, 0);
	      
	    }
#else
	    err_debug(DB_STARTUP,("nuoms","calling oms_BranchInPlace directly"));
	    (void)oms_real_BranchInPlace(global,local,0,0);
#endif
	  }
	  /* force reinitialization on next loop */
	  matroid_eq->initialized=0;
	  RestoreState(local);
	}
      }
      na_destroy(local);
    } while (parse_input(params));
    output_finish(params);
  }
 
#ifdef USE_ZRAM
    fprintf(stderr,"Processors: %d\n",z_numprocs);
    z_DumpMemUsage();
#endif

  /*  shutdown(global,HERE,"search completed"); */
  prof_dump(params->profile_level);
  na_destroy(global);
  return(0);
}
/**
   \brief This is the routine invoked by ZRAM. It does not return 
   the state, because there is no where to return it to
*/
bool oms_BranchInPlace(na_node_p global,  na_node_p state,
			  ZRAM_ARGS_PROTO){
  na_fixup(global);
  na_fixup(state);

  (void)oms_real_BranchInPlace(global,state,ZRAM_ARGS);
  
  na_destroy(state); 
  return(0);
}

bool nuoms_success(na_node_p global,na_node_p state, params_p params){
  params->success_count++;
  prof_event(PROF_SUCCESS);
  (void) output_state(global,state);
  if (ERR_DEBUG_MASK & DB_BACKTRACK){
    sarray_p  chirotope=na_access(state,SLOT_chirotope);
    char *buf= string_pm(chirotope,sign2char);
    err_warning("success","chi=%s",buf);
    free(buf);
  };

  if (params->max_results>=1 && params->max_results <= params->success_count){
    err_debug(DB_BACKTRACK,("nuoms","Skipping rest of search as per request"));
    return FALSE;
  }
  return TRUE;
}
/** 
 * Do the actual work for Branch in place, but don't fixup up state 
 * blocks from (possible) network transmission.
 * 
 * @param global Global state
 * @param local Local state
 * 
 * @return true if search should continue
 */
bool oms_real_BranchInPlace(na_node_p global,  
				    na_node_p local,
				    ZRAM_ARGS_PROTO){

  params_p params=na_access(global,SLOT_params);  
  int_stack_p bt_stack=na_access(local,SLOT_bt_stack);
  lparams_p lparams=na_access(local,SLOT_lparams);
  bool explore=TRUE;

  if (params->max_depth>=0 && lparams->depth>=params->max_depth){
    output_state(global,local);
    prof_event(PROF_DEPTH_LIMIT);
    return FALSE;
  }

  SaveState(local);

  if (!nvc_push_choices(global,local,bt_stack)){
    /* all signs are forced by the input */
    (void)nuoms_success(global,local,params);
    err_debug(DB_BACKTRACK,("nuoms","all chirotope signs forced by input"));
    return false;
  }    

  while(!istack_empty(bt_stack)){
      nu_signed_index_t choice;

      choice=istack_pop(bt_stack);
      if (choice==NSI_MARKER){
	RestoreState(local);
	explore=FALSE;
      } else {
	explore=TRUE;
      }


      if(explore) {
	array_index_t basis=nsi_index(choice);
	nu_sign_t sign=nsi_sign(choice);

	if (OPT_detailed_profile)
	  prof_event(PROF_NODE);

	lparams->subtreesize++;

	SaveState(local);
	explore=SetAndConsequences(global,local,basis,sign);
	if (explore){
	  if (!nvc_push_choices(global,local,bt_stack)){
	    /* note: this logic can and should be simplified; unlike oms, 
	     we do not call real_BranchInPlace recursively */
	    bool keep_looking=nuoms_success(global,local,params);
	    RestoreState(local);
	    if (!keep_looking)
	      return FALSE;
	  }
	} else {
	  RestoreState(local);
	}
      }
  }


  return TRUE;
}


/** 
 * Test single variable setting for consistency.


 This function has two interrelated jobs: 
 - test if the suggested
 setting is consistent (i.e. ok with pl\"{u}cker, boundary, and diamater)
 - Find variable settings forced by this one.

 * 
 * @param state local state
 * @param global global state
 * @param basis which variable to set
 * @param sigma what sign to set it to
 * @param beta0 root basis index (-1 = unknown)
 *
 * @return True if that setting is consistent (as far as we can tell)
 */
bool SetAndConsequences(  na_node_p	global,
			  na_node_p	state,
			   array_index_t	basis, 
			  nu_sign_t	sigma){

  sarray_p chirotope;

  be_block_p PE;		/* %%Pl\"ucker%% equations */
  sa_array_p pe_state;
  params_p params;
  
  nu_sign_t old_sign;
  nu_sign_set_p forced; /* forced variables */
  lparams_p lparams=na_access(state,SLOT_lparams);

  NODECHECK(state);

  chirotope=na_access(state,SLOT_chirotope);
  pe_state=na_access(state,SLOT_pe_state);
  
  PE=na_access(global,SLOT_pe);

  params=na_access(global,SLOT_params);
  
  /* note that this check is mostly unnecessary, except during startup */
  old_sign=sa_read(chirotope,basis);

  if (sign_subset(old_sign,sigma))
    return TRUE;

  if (!sign_subset(sigma,old_sign))
    ERR_RETURN(FALSE);

  forced=nss_new(sa_size(chirotope));

  err_debug(DB_BACKTRACK,(HERE,"initializing forced with %d->%c",basis,
	    sign2char(sigma)));

  (void)nss_add(forced,basis,sigma);

  while (!nss_empty(forced)){
    int choice;
    int bindex;
    nu_sign_t sign;

    if (ERR_DEBUG_MASK&DB_BACKTRACK) {
      fprintf(stderr,"forced= ");
      nss_dump(forced);
      fprintf(stderr,"\n");
    }

    choice=nss_deq(forced);
    bindex=nsi_index(choice);
    sign=nsi_sign(choice);

    err_debug(DB_BACKTRACK,(HERE,"setting %d->%s from forced\n",bindex,sign_name[sign]));


    sa_set(chirotope,bindex,sign);

    if (!params->no_boundary && !nb_traverse(bindex,
		     state,global, 
		     forced,NULL)){
      err_debug(DB_BACKTRACK,(HERE,"boundary eqns failed")); 
      nss_destroy(forced); 
      ERR_RETURN(FALSE);  
    }

    if (!params->no_convexity && !nc_traverse(bindex,
		     state,global, 
		     forced,NULL)){
      err_debug(DB_BACKTRACK,(HERE,"convexity eqns failed")); 
      nss_destroy(forced); 
      ERR_RETURN(FALSE);  
    }

    if ((lparams->root_basis >=0) && (!params->no_matroid) && !me_traverse(bindex,state,global,forced)){
      err_debug(DB_BACKTRACK,(HERE,"matroid eqns failed"));
      nss_destroy(forced); 
      ERR_RETURN(FALSE);  
    }

    if (!params->no_grassman_pluecker && !npl_traverse(bindex, PE,
			      forced,chirotope,pe_state)){
      err_debug(DB_BACKTRACK,(HERE,"pluecker eqns failed"));
      nss_destroy(forced);
      ERR_RETURN(FALSE);
    }
  }
  nss_destroy(forced);
  return TRUE;
}

#if (USE_ZRAM)
/**
   ZRAM initialization.
 This function is called once on every processing element. 
*/
void initproc(void)
{

  
  z_signal_Init();

  z_bt1_Init();
  z_bt2_Init();
  z_check_Init();
  oms_NodeTag = z_InstallClass(oms_Node, 0, 0, 0);
  oms_GlobalTag = z_InstallClass(oms_Global, 0, 0, 0);
  assert(oms_GlobalTag>0);
  oms_procset.bt_BranchInPlace = (z_bt_BranchProc)oms_BranchInPlace;
  oms_procsetnum = z_InstallProcSet((z_ProcSetP) &oms_procset);
  
#ifdef USE_PROFILE
  prof_zram_init();
#endif /* USE_PROFILE */

} /* initproc */

#endif

/*
extern bool SetSign(na_node_p local, na_node_p global,
		    array_index_t bindex, nu_sign_t sign){
  sarray_p chirotope;
  oms_assert(sign*sign==1);
  oms_debug(3,("SetSign","%d %d",bindex,sign));
  chirotope=na_access(local,SLOT_chirotope);
  sa_set(chirotope,bindex,sign);
}
*/

bool nuoms_read_boundary_hook(na_node_p global, na_node_p local, 
			    int b, int sigma)
{
  be_block_p bdy_eq=na_access(global,SLOT_bdy_eq);
  array_index_t    container=be_bindex(bdy_eq,b,0);
  nu_sign_set_p choice_Q=na_access(local,SLOT_choice_Q);
  
  err_debug(DB_STARTUP,("oms","saving choice %d",container));
  (void)nss_add(choice_Q,container,SIGN_ANY); /* enqueue some basis contining b */
  return TRUE;

}

static na_node_p nuoms_alloc_local(na_node_p global,na_node_p local){
 
  params_p params=na_access(global,SLOT_params);

  int n=params->n;
  int r=params->r;
  int num_bases=binomial(n,r);
  int num_eqns=binomial(n,r+2)*binomial(r+2,4); 

  /* estimate */
  int pe_saves=binomial(n-r,2)*binomial(r,2)*r;
  local=sa_nalloc(local,SLOT_pe_state, num_eqns, pe_saves);
  local=nss_nalloc(local,SLOT_choice_Q,num_bases);
  local=istack_nalloc(local,SLOT_bt_stack,num_bases*4);
  na_fixup(local);

  {
    sarray_p   pe_state=na_access(local,SLOT_pe_state);
    int i;

    for(i=0; i<sa_size(pe_state); i++)
      sa_init(pe_state,i,PL_ROW_UNKNOWN);
  }
  {
    lparams_p lparams=na_access(local,SLOT_lparams);

    lparams->depth=0;
    lparams->subtreesize=0;
  }

  /* what is this about? 
  na_destroy(local);
  */

  return local;
}

void SaveState(search_state state){


  lparams_p lparams=na_access(state,SLOT_lparams);
  lparams->depth++;
  if (ERR_DEBUG_MASK & DB_BACKTRACK){
    sarray_p  chirotope=na_access(state,SLOT_chirotope);
    char *buf= string_pm(chirotope,sign2char);
    err_warning("SaveState","depth= %d, chi=%s",lparams->depth,buf);
    free(buf);
  };
  rg_save(na_access(state,SLOT_ridge_graph)); 
  sa_save(na_access(state,SLOT_chirotope));
  sa_save(na_access(state,SLOT_conv_neg_mask));
  sa_save(na_access(state,SLOT_conv_pos_mask));
  sa_save(na_access(state,SLOT_conv_zero_mask));
  sa_save(na_access(state,SLOT_facet_orientation));
  sa_save(na_access(state,SLOT_matroid_nonzero_mask));
  sa_save(na_access(state,SLOT_matroid_zero_mask));
  sa_save(na_access(state,SLOT_neg_mask));
  sa_save(na_access(state,SLOT_pe_state));
  sa_save(na_access(state,SLOT_pos_mask));
  sa_save(na_access(state,SLOT_zero_mask));


  sa_save(na_access(state,SLOT_boundary));
}


extern void RestoreState(search_state state){

  lparams_p lparams;

  NODECHECK(state);

  lparams=na_access(state,SLOT_lparams);


  lparams->depth--;
  err_debug(DB_BACKTRACK,("RestoreState","depth= %d",lparams->depth));
 
  err_assert(lparams->depth>=0);


  rg_restore(na_access(state,SLOT_ridge_graph));
  sa_restore(na_access(state,SLOT_boundary));
  sa_restore(na_access(state,SLOT_chirotope));
  sa_restore(na_access(state,SLOT_conv_neg_mask));
  sa_restore(na_access(state,SLOT_conv_pos_mask));
  sa_restore(na_access(state,SLOT_conv_zero_mask));
  sa_restore(na_access(state,SLOT_facet_orientation));
  sa_restore(na_access(state,SLOT_matroid_nonzero_mask));
  sa_restore(na_access(state,SLOT_matroid_zero_mask));
  sa_restore(na_access(state,SLOT_neg_mask));
  sa_restore(na_access(state,SLOT_pe_state));
  sa_restore(na_access(state,SLOT_pos_mask));
  sa_restore(na_access(state,SLOT_zero_mask));

 
}
