/* z_enumps.c (ZRAM: Enumeration: PostScript output)
 * Ambros Marzetta
 *
 * $Source$
 * $Revision$
 * $Date$
 * $Log$
 * Revision 1.2  2000/11/15 16:14:52  bremner
 * committing patched version of zram source
 *
 * Revision 1.4  1998/04/15  12:17:30  marzetta
 * Fixed a bug in the multi-page code
 *
 * Revision 1.3  1998/03/27  16:47:50  marzetta
 * We can now print more than one page.
 *
 * Revision 1.2  1997/04/18  14:59:49  marzetta
 * Added #ifdef ENCAPSULATED. Define ENCAPSULATED in order to generate EPS
 * files instead of PostScript jobs.
 *
 * Revision 1.1  1996/06/03  19:01:41  marzetta
 * Initial revision
 *
 */


#include <unistd.h>
#include <fcntl.h>
#include "z_enumps.h"
#include "z_io.h"
#include <math.h>

static int fd;
static int fd_elem_count;
static boolean fd_opened;

void z_enumps_Open(const char *name)
{
  fd = open((char *)name, O_RDWR|O_CREAT|O_APPEND|O_TRUNC, 0666);
  if (fd < 0) perror("open");
  fd_opened = fd >= 0;
}

void z_Save(z_NodeP global, z_NodeP node)
/* Save node to the static file fd. Increment static fd_elem_count. */
{
  assert(fd_opened);
  (void)write(fd, node, node->len);
  assert(node->children == 0);
  fd_elem_count ++;
} /* cis_Save */

z_NodeP z_Load(void)
{
  struct z_Node header;
  z_NodeP node;

  assert(fd_opened);
  (void)read(fd, &header, sizeof(header));
  node = z_NewNode(header.tag, header.len);
  (void)read(fd, (void *)((long)node + sizeof(header)),
	     header.len - sizeof(header));
  return node;
} /* z_Load */

void z_enumps_Rewind(int32 *elem_count)
{
  assert(fd_opened);
  (void) lseek(fd, 0, SEEK_SET);
  *elem_count = fd_elem_count;
} /* z_enumps_Rewind */

#ifdef ENCAPSULATED
#define PAPER_BORDER	20
#define PS_SIZE_X 	427	/* EPS: 427 pt = 150 mm */
                                /* writable area in pt */
#define PS_SIZE_Y 	300	/* EPS: 300 pt = 105 mm */
#else /* not ENCAPSULATED */
#define PAPER_NAME	"A4"
#define PAPER_SIZE_X	595	/* A4: 595 pt = 209 mm */
#define PAPER_SIZE_Y	842	/* A4: 842 pt = 297 mm */
#define PAPER_BORDER	45	/*      45 pt =  15 mm */
#define PS_SIZE_X 	(PAPER_SIZE_X - 2 * PAPER_BORDER)
                                /* writable area in pt (510,750)*/
#define PS_SIZE_Y 	(PAPER_SIZE_Y - 2 * PAPER_BORDER)
#endif /* ENCAPSULATED */

#define BORDER_FRACTION (1/36.0) /* distance from border outside */
#define MAX_ELEM	600	/* pictures per page */
#define LINEWIDTH	0.2	/* line width of frame */
#define HLW		((int)ceil(LINEWIDTH/2)) /* to compute the bounding box */

float z_enumps_ratio = 1;	/* height / width */
boolean z_enumps_frame = 1;	/* draw a rectangular frame */
float z_enumps_borderFraction = 1/36.0; /* space between frames */

void z_PrintPS(z_NodeP global, FILE *psfile,
	       void (*AppDepHeader) (z_NodeP, FILE *, float, float),
	       void (*AppDepPicture) (z_NodeP, z_NodeP, FILE *, float, float))
{
  /* Print all nodes saved by z_Save to the PostScript file psfile.
     z_enumps_ratio = height / width of a node.
     Call AppDepPicture with unscaled coordinates, but origin at lower
     left of its rectangle. */
  int num_x, num_y, i, page;
  float xtrans, ytrans, xsize, ysize;
  z_NodeP node;
  int32 elem_count, elem_per_page, elem_on_this_page;

  z_enumps_Rewind(&elem_count);
  elem_per_page = elem_count;
  if (elem_per_page > MAX_ELEM) elem_per_page = MAX_ELEM;
  { /* Compute number of elements in x and y direction. */
    num_x = ceil(sqrt(elem_per_page *
		      ((float)PS_SIZE_X / (float)PS_SIZE_Y * z_enumps_ratio)));
    assert(num_x > 0);
    num_y = (elem_per_page + num_x - 1) / num_x;
  }

  xtrans = (float)PS_SIZE_X / num_x; /* translation between frames */
  ytrans = (float)PS_SIZE_Y / num_y;
  if (ytrans > z_enumps_ratio * xtrans) ytrans = z_enumps_ratio * xtrans;
  if (xtrans > ytrans / z_enumps_ratio) xtrans = ytrans / z_enumps_ratio;
  xsize = xtrans / (1 + 2 * z_enumps_borderFraction); /* size of frame in points */
  ysize = ytrans / (1 + 2 * z_enumps_borderFraction);
  
  { /* Print PostScript File Header */
#ifdef ENCAPSULATED
    fprintf(psfile, "%%!PS-Adobe-3.0\n"
	    "%%%%Creator: (ZRAM:%s)\n"
	    "%%%%DocumentFonts:\n"
	    "%%%%BoundingBox: %d %d %d %d\n"
	    "%%%%Orientation: Portrait\n"
	    "%%%%EndComments\n"
	    "\n", __FILE__,	/* creator */
	    PAPER_BORDER - HLW,	/* bb: left */
	    PAPER_BORDER + PS_SIZE_Y - (int)(num_y*ytrans) - HLW,/*bb: bottom*/
	    PAPER_BORDER + (int)((num_x-1)*xtrans+xsize) + HLW,	/* bb: right */
	    PAPER_BORDER + PS_SIZE_Y +(int)(-ytrans+ysize) + HLW); /* bb: top*/
#else /* not ENCAPSULATED */
    fprintf(psfile, "%%!PS-Adobe-3.0\n"
	    "%%%%Creator: (ZRAM:%s)\n"
	    "%%%%Pages: %d\n"
	    "%%%%DocumentFonts:\n"
	    "%%%%BoundingBox: %d %d %d %d\n"
	    "%%%%Orientation: Portrait\n"
	    "%%%%DocumentMedia: %s %d %d 0 () ()\n"
	    "%%%%EndComments\n"
	    "\n", __FILE__,	/* creator */
	    (elem_count-1) / elem_per_page + 1,	/* pages */
	    PAPER_BORDER - HLW,	/* bb: left */
	    PAPER_BORDER + PS_SIZE_Y - (int)(num_y*ytrans) - HLW,/*bb: bottom*/
	    PAPER_BORDER + (int)((num_x-1)*xtrans+xsize) + HLW,	/* bb: right */
	    PAPER_BORDER + PS_SIZE_Y +(int)(-ytrans+ysize) + HLW, /* bb: top */
	    PAPER_NAME, PAPER_SIZE_X, PAPER_SIZE_Y);
#endif /* ENCAPSULATED */

    /* define xtrans and ytrans */
    fprintf(psfile, "/xtrans %1.2f def /ytrans %1.2f def\n",
	    xtrans, ytrans);
    /* define xsize and ysize */
    fprintf(psfile, "/xsize %1.2f def /ysize %1.2f def\n",
	    xsize, ysize);
    /* move origin onto paper (bottom left of top left element) */
    fprintf(psfile, "%d %d ytrans sub translate\n",
	    PAPER_BORDER, PAPER_BORDER + PS_SIZE_Y);
    /* define frame */
    if (z_enumps_frame) {
      fprintf(psfile, "/frame {\n");
      fprintf(psfile, "  %1.2f setlinewidth\n", LINEWIDTH);
      fprintf(psfile, "  newpath 0 0 moveto xsize 0 rlineto 0 ysize rlineto\n"
	      "  0 xsize sub 0 rlineto closepath stroke\n");
      fprintf(psfile, "} def\n");
    }
    /* set line width */
    fprintf(psfile, "%1.2f setlinewidth\n", LINEWIDTH);
    /* define gre and grel (end of line) */
    fprintf(psfile, "/gre {grestore xtrans 0 translate} def\n");
    fprintf(psfile, "/grel {grestore 0 xtrans %d mul sub 0 ytrans sub translate} def\n",
	    num_x - 1);
    /* application dependent header */
    fprintf(psfile, "%% Start of application dependent header\n");
    AppDepHeader(global, psfile, xsize, ysize);
    fprintf(psfile, "%%%%EndProlog\n\n");
  }

  for (page=1; page<=(elem_count-1) / elem_per_page + 1; page++) {
    fprintf(psfile, "\n%%%%Page: %d %d\n", page, page);
    fprintf(psfile, "/saveobj save def\n");
    if (page * elem_per_page >= elem_count) {
      elem_on_this_page = (elem_count-1) % elem_per_page + 1; /* last page */
    } else
	elem_on_this_page = elem_per_page; /* not last page */
    for (i=0; i<elem_on_this_page; i++) { /* make picture number i */

      node = z_Load();
      fprintf(psfile, "%sgsave\n", z_enumps_frame ? "frame " : "");
      AppDepPicture(global, node, psfile, xsize, ysize);
      z_FreeNode(node);
      if ((i+1) % num_x != 0) {
	fprintf(psfile, "gre\n");
      } else
	  fprintf(psfile, "grel\n");
    } /* for picture number i */
    fprintf(psfile, "showpage\n");
    fprintf(psfile, "saveobj restore\n");
  }
  
  fprintf(psfile, "\n%%%%Trailer\n");
} /* z_PrintPS */

void z_enumps_init(void)
{
}
